package UroIrmPackage;

import ij.*;
import ij.gui.*;
import ij.text.*;
import ij.plugin.*;
import ij.process.*;
import ij.measure.*;
import java.awt.*;
import java.util.*;

/** same class as "ij.plugin.filter.Aalyzer", but the preferences are found with 
PrefSaveResult and not Pref
*/
public class UroIrmSaveResult implements Measurements,UroIrmMessage {
	private static TextPanel textPanel;
	private ImagePlus imp;
	private HeaderDicom header ;

	private String ACQTPS = "acq tps(mm,ss)";	
	private String ACQTPSSEC = "acq (sec)";
	private String NSLICE = "n";	
	public static String MARK_DATA = "data : ";

	private String Title;
	private ResultsTable rt;
	private LinkedList evaluationTable;
	private int measurements;
	private StringBuffer min,max,mean,sd;

	private static final int[] list = {AREA,MEAN,STD_DEV,MODE,MIN_MAX,
		CENTROID,CENTER_OF_MASS,PERIMETER,RECT,ELLIPSE,LIMIT,LABELS,INVERT_Y};

	private static final int UNDEFINED=0,AREAS=1,LENGTHS=2,ANGLES=3,MARK_AND_COUNT=4;
	private static int mode = UNDEFINED;
	private static final String MEASUREMENTS = "saveresult";
	private static final String MARK_WIDTH = "mark.width";
	private static final String PRECISION = "precision";
	//private static int counter;
	private static boolean unsavedMeasurements;
	public static Color darkBlue = new Color(0,0,160);
	private static int systemMeasurements ;
	public static int markWidth;
	public static int precision ;
	private static float[] umeans = new float[MAX_STANDARDS];
	private static ResultsTable systemRT = new ResultsTable();

	public UroIrmSaveResult(ImagePlus Psource, TextPanel Tsource) {
		imp = Psource;
		header = new HeaderDicom(imp);
		PrefSaveResult.load();
		textPanel = Tsource;
		rt = systemRT;
		rt.setPrecision(precision);
		Title = "saveResult";
		systemMeasurements = PrefSaveResult.getSetInt(MEASUREMENTS,MEAN+LABELS);
		markWidth = PrefSaveResult.getSetInt(MARK_WIDTH,3);
		precision = PrefSaveResult.getSetInt(PRECISION,3);
		measurements = systemMeasurements;
		evaluationTable = new LinkedList();
	}

	public void setTextPanel(TextPanel Tsource) {
		textPanel = Tsource;
	}

	public void setTitle(String txt){
		Title=txt;
	}

	public void doSetDialog() {
		GenericDialog gd = new GenericDialog("Set Measurements", IJ.getInstance());
		String[] labels = new String[10];
		boolean[] states = new boolean[10];
		labels[0]="Area"; states[0]=(measurements&AREA)!=0;
		labels[1]="Mean Gray Value"; states[1]=(measurements&MEAN)!=0;
		labels[2]="Standard Deviation"; states[2]=(measurements&STD_DEV)!=0;
		labels[3]="Modal Gray Value"; states[3]=(measurements&MODE)!=0;
		labels[4]="Min & Max Gray Value"; states[4]=(measurements&MIN_MAX)!=0;
		labels[5]="Centroid"; states[5]=(measurements&CENTROID)!=0;
		labels[6]="Center of Mass"; states[6]=(measurements&CENTER_OF_MASS)!=0;
		labels[7]="Perimeter"; states[7]=(measurements&PERIMETER)!=0;
		labels[8]="Bounding Rectangle"; states[8]=(measurements&RECT)!=0;
		labels[9]="Fit Ellipse"; states[9]=(measurements&ELLIPSE)!=0;
		gd.addCheckboxGroup(5, 2, labels, states);
		labels = new String[3];
		states = new boolean[3];
		labels[0]="Limit to Threshold"; states[0]=(measurements&LIMIT)!=0;
		labels[1]="Stack Header"; states[1]=(measurements&LABELS)!=0;
		labels[2]="Invert Y Coordinates"; states[2]=(measurements&INVERT_Y)!=0;
		gd.addCheckboxGroup(2, 2, labels, states);
		gd.addMessage("");
		gd.addNumericField("Decimal Places:", precision, 0);
		gd.showDialog();
		if (gd.wasCanceled())
			return;
		setOptions(gd);
		int prec = (int)gd.getNextNumber();
		if (prec>=0 && prec<=8 && prec!=precision) {
			precision = prec;
			rt.setPrecision(precision);
			if (mode==AREAS) {
				textPanel.setColumnHeadings("");
				updateHeadings();
			}
		}
		PrefSaveResult.setInt(MEASUREMENTS,measurements);
		PrefSaveResult.setInt(MARK_WIDTH,markWidth);
		PrefSaveResult.setInt(PRECISION,precision);
	}
	
	void clearWorksheet() {
		resetCounter();
	}
	 
	void setOptions(GenericDialog gd) {
		int oldMeasurements = measurements;
		int previous = 0;
		boolean b = false;
		for (int i=0; i<list.length; i++) {
			//if (list[i]!=previous)
				b = gd.getNextBoolean();
			previous = list[i];
			if (b)
				measurements |= list[i];
			else
				measurements &= ~list[i];
		}
		if ((oldMeasurements&(~LIMIT))!=(measurements&(~LIMIT)))
			mode = UNDEFINED;
		if ((measurements&LABELS)==0)
			systemRT.disableRowLabels();
	}
	
	public void measure() {
		int currentSlice = imp.getCurrentSlice();
		header.setSlice(currentSlice);
		String status = header.getTagValue(HeaderDicom.ETAT);
		
		if ( status != null && status.equals(HeaderDicom.DISABLE) )
			return;

		ImageStatistics stats = imp.getStatistics(measurements);
		saveResults(stats, imp.getRoi());
	}

/** save the evaluations made */
	void addEvaluation(float[] point1,float[] point2,float res){
		float[][] element = new float[3][];
		element[0]=point1;
		element[1]=point2;
		element[2]= new float[1] ; element[2][0] = res;
		evaluationTable.add(element);
	}

	/** Saves the measurements specified in the "Set Measurements" dialog,
		or by calling setMeasurments(), in the system results table.
	*/
	public void saveResults(ImageStatistics stats, Roi roi) {		
		incrementCounter();
		int counter = rt.getCounter();
		if (counter<=MAX_STANDARDS) {
			if (umeans==null) umeans = new float[MAX_STANDARDS];
			umeans[counter-1] = (float)stats.umean;
		}
		int currentSlice = imp.getCurrentSlice();
		header.setSlice(currentSlice);
		String S_acqTime = header.getTagValue(HeaderDicom.ACQTIME);
		UroIrmTime data = new UroIrmTime(S_acqTime);

		rt.addValue(NSLICE,currentSlice);      
		rt.addLabel(ACQTPS, data.toString());
		rt.addValue(ACQTPSSEC,data.getValue());
		
		if ((measurements&AREA)!=0) rt.addValue(ResultsTable.AREA,stats.area);
		if ((measurements&MEAN)!=0) rt.addValue(ResultsTable.MEAN,stats.mean);
		if ((measurements&STD_DEV)!=0) rt.addValue(ResultsTable.STD_DEV,stats.stdDev);
		if ((measurements&MODE)!=0) rt.addValue(ResultsTable.MODE, stats.dmode);
		if ((measurements&MIN_MAX)!=0) {
			rt.addValue(ResultsTable.MIN,stats.min);
			rt.addValue(ResultsTable.MAX,stats.max);
		}
		if ((measurements&CENTROID)!=0) {
			rt.addValue(ResultsTable.X_CENTROID,stats.xCentroid);
			rt.addValue(ResultsTable.Y_CENTROID,updateY(stats.yCentroid));
		}
		if ((measurements&CENTER_OF_MASS)!=0) {
			rt.addValue(ResultsTable.X_CENTER_OF_MASS,stats.xCenterOfMass);
			rt.addValue(ResultsTable.Y_CENTER_OF_MASS,updateY(stats.yCenterOfMass));
		}
		if ((measurements&PERIMETER)!=0) {
			double perimeter;
			if (roi!=null)
				perimeter = roi.getLength();
			else
				perimeter = 0.0;
			rt.addValue(ResultsTable.PERIMETER,perimeter);
		}
		if ((measurements&RECT)!=0) {
			rt.addValue(ResultsTable.ROI_X,stats.roiX);
			rt.addValue(ResultsTable.ROI_Y,updateY2(stats.roiY));
			rt.addValue(ResultsTable.ROI_WIDTH,stats.roiWidth);
			rt.addValue(ResultsTable.ROI_HEIGHT,stats.roiHeight);
		}
		if ((measurements&ELLIPSE)!=0) {
			rt.addValue(ResultsTable.MAJOR,stats.major);
			rt.addValue(ResultsTable.MINOR,stats.minor);
			rt.addValue(ResultsTable.ANGLE,stats.angle);
		}
	}


	// Update centroid and center of mass y-coordinate
	// based on value "Invert Y Coordinates" flag
	double updateY(double y) {
		if (imp==null)
			return y;
		else {
			if ((systemMeasurements&INVERT_Y)!=0) {
				Calibration cal = imp.getCalibration();
				y = imp.getHeight()*cal.pixelHeight-y;
			}
			return y;
		}
	}

	// Update bounding rectangle y-coordinate based
	// on value "Invert Y Coordinates" flag
	double updateY2(double y) {
		if (imp==null)
			return y;
		else {
			if ((systemMeasurements&INVERT_Y)!=0) {
				Calibration cal = imp.getCalibration();
				y = imp.getHeight()*cal.pixelHeight-y-cal.pixelHeight;
			}
			return y;
		}
	}



	/** Writes the last row in the results table to the ImageJ window. */
	public void displayResults() {
		int counter = rt.getCounter();
		StringTokenizer headings = new StringTokenizer(rt.getColumnHeadings(),"\t");
		if ( evaluationTable.size()!= 0 && headings.countTokens() < 7 ) {
			int nb = 7-headings.countTokens() ;
			for (int i =0;i<nb;i++)
				rt.addValue("aux" + String.valueOf(i),0.0);
		}
		textPanel.setColumnHeadings(rt.getColumnHeadings());		
		if ( (measurements & LABELS)!=0){
			textPanel.append((Title.equals(""))?ROI_NAME + NONE : ROI_NAME+Title);
			String text = imp.getTitle();
			int index = Header.STACK.length();
			if (index < text.length() ) 
				text = INFORMATION + text.substring(index);	
			else if ( index == text.length() )
				text = INFORMATION + NONE;	
			textPanel.append(text);
			textPanel.append(MARK_DATA);
		}
		for ( int i =0; i< counter;i++){ 
			textPanel.append(rt.getRowAsString(i));
		}
		if ( evaluationTable.size() == 0)
			return;
		textPanel.append(msg_comm_result_evaluation + "  ( n = image number)");
		textPanel.append(	" \t first point \t \t \t second point \t \t  result" );
		textPanel.append(" x1 \t y1 \t n \t x2 \t y2 \t n  \t result"); 
		for (ListIterator iter = evaluationTable.listIterator(0);iter.hasNext();){
			float[][] element = (float[][])iter.next();
			String text = f2s(element[0][0])+ "\t" + f2s(element[0][1])+ "\t" +f2s(element[0][2])+ "\t";
			text += f2s(element[1][0])+ "\t" + f2s(element[1][1])+ "\t" +f2s(element[1][2])+ "\t";
			text += f2s(element[2][0]); 
			textPanel.append(text);
		}
	}
	
	private String f2s (float f) { return String.valueOf(f);}

	public void update(){
		umeans = null;
		systemRT.reset();
		unsavedMeasurements = false;
		textPanel.selectAll();
		textPanel.clearSelection();

		updateHeadings();
		if ((measurements&LABELS)!=0 ) 
			addLabel(imp.getTitle());
	
	}

	private void addLabel( String s){
		int nb = textPanel.getLineCount();
		String save = "";
		if ( nb < 0 )			
			return;
		String first = textPanel.getLine(0);
		if (first.equals(MARK_DATA)) 
			save += s + "\n" + first + "\n";
		else {
			if (first.indexOf(s) <0 )
				save += first + s +"\n";
			else 
				save += first+ "\n";
		}		
		for ( int line = 1 ; line < nb -1 ;line ++){
			save += textPanel.getLine(line)+ " \n " ;
		}
		save += textPanel.getLine(nb-1);
		textPanel.selectAll();
		textPanel.clearSelection();
		textPanel.append(save);
	}

	public void updateHeadings() {
		String worksheetHeadings = textPanel.getColumnHeadings();		
		String tableHeadings = rt.getColumnHeadings();		
		if (worksheetHeadings.equals(tableHeadings))
			return;
		textPanel.setColumnHeadings(tableHeadings);
		int n = rt.getCounter();
		textPanel.append(MARK_DATA );
		if (n>0) {
			StringBuffer sb = new StringBuffer(n*tableHeadings.length());
			for (int i=0; i<n; i++)
				sb.append(rt.getRowAsString(i)+"\n");
			textPanel.append(new String(sb));
		}
	}

	/** Converts a number to a formatted string with a tab at the end. */
	public String n(double n) {
		String s;
		if (Math.round(n)==n)
			s = IJ.d2s(n,0);
		else
			s = IJ.d2s(n,precision);
		return s+"\t";
	}
		
	void incrementCounter() {
		//counter++;
		if (rt==null) rt = systemRT;
		rt.incrementCounter();
		unsavedMeasurements = true;
	}
	
	public void summarize() {
		rt = systemRT;
		if (rt.getCounter()<1){
			IJ.error(msg_err_no_measure);	
			return;
		}
		measurements = systemMeasurements;
		min = new StringBuffer(100);
		max = new StringBuffer(100);
		mean = new StringBuffer(100);
		sd = new StringBuffer(100);
		min.append("Min\t");
		max.append("Max\t");
		mean.append("Mean\t");
		sd.append("SD\t");
		if ((measurements&LABELS)!=0) {
			min.append("\t");
			max.append("\t");
			mean.append("\t");
			sd.append("\t");
		}
		if (mode==MARK_AND_COUNT) 
			summarizePoints(rt);
		else if (mode==LENGTHS) 
			add2(rt.getColumnIndex("Length"));
		else
			summarizeAreas();
		String worksheetHeadings = textPanel.getColumnHeadings();		
		if (worksheetHeadings.equals(""))
			textPanel.setColumnHeadings(rt.getColumnHeadings());
	
		textPanel.append("");
		
		textPanel.append(new String(mean));		
		textPanel.append(new String(sd));		
		textPanel.append(new String(min));		
		textPanel.append(new String(max));
		textPanel.append("");		
		mean = null;		
		sd = null;		
		min = null;		
		max = null;		
	}
	
	void summarizePoints(ResultsTable rt) {
		add2(rt.getColumnIndex("X"));
		add2(rt.getColumnIndex("Y"));
		add2(rt.getColumnIndex("Value"));
	}

	void summarizeAreas() {
		if ((measurements&AREA)!=0) add2(ResultsTable.AREA);
		if ((measurements&MEAN)!=0) add2(ResultsTable.MEAN);
		if ((measurements&STD_DEV)!=0) add2(ResultsTable.STD_DEV);
		if ((measurements&MODE)!=0) add2(ResultsTable.MODE);
		if ((measurements&MIN_MAX)!=0) {
			add2(ResultsTable.MIN);
			add2(ResultsTable.MAX);
		}
		if ((measurements&CENTROID)!=0) {
			add2(ResultsTable.X_CENTROID);
			add2(ResultsTable.Y_CENTROID);
		}
		if ((measurements&CENTER_OF_MASS)!=0) {
			add2(ResultsTable.X_CENTER_OF_MASS);
			add2(ResultsTable.Y_CENTER_OF_MASS);
		}
		if ((measurements&PERIMETER)!=0)
			add2(ResultsTable.PERIMETER);
		if ((measurements&RECT)!=0) {
			add2(ResultsTable.ROI_X);
			add2(ResultsTable.ROI_Y);
			add2(ResultsTable.ROI_WIDTH);
			add2(ResultsTable.ROI_HEIGHT);
		}
		if ((measurements&ELLIPSE)!=0) {
			add2(ResultsTable.MAJOR);
			add2(ResultsTable.MINOR);
			add2(ResultsTable.ANGLE);
		}
	}

	private void add2(int column) {
		float[] c = column>=0?rt.getColumn(column):null;
		if (c!=null) {
			ImageProcessor ip = new FloatProcessor(c.length, 1, c, null);
			if (ip==null)
				return;
			ImageStatistics stats = new FloatStatistics(ip);
			if (stats==null)
				return;
			mean.append(n(stats.mean));
			min.append(n(stats.min));
			max.append(n(stats.max));
			sd.append(n(stats.stdDev));
		} else {
			mean.append("-\t");
			min.append("-\t");
			max.append("-\t");
			sd.append("-\t");
		}
	}

	/** Returns the current measurement count. */
	public static int getCounter() {
		return systemRT.getCounter();
	}

	/** Sets the measurement counter to zero. Displays a dialog that
	    allows the user to save any existing measurements. Returns
	    false if the user cancels the dialog.
	*/
	public synchronized static boolean resetCounter() {
		int lineCount = textPanel.getLineCount();
		int counter = systemRT.getCounter();
		if (counter>0 && lineCount>0 && unsavedMeasurements && !IJ.macroRunning()) {
			SaveChangesDialog d = new SaveChangesDialog(IJ.getInstance(), "Save "+counter+" measurements?");
			if (d.cancelPressed())
				return false;
			else if (d.savePressed())
				new MeasurementsWriter().run("");
		}
		umeans = null;
		systemRT.reset();
		unsavedMeasurements = false;
		textPanel.selectAll();
		textPanel.clearSelection();
		return true;
	}
	
	public static void setSaved() {
		unsavedMeasurements = false;
	}
	
	// Returns the measurements defined in the Set Measurements dialog. */
	public static int getMeasurements() {
		return systemMeasurements;
	}

	// Sets the system-wide measurements. */
	public static void setMeasurements(int measurements) {
		systemMeasurements = measurements;
	}

	/** Returns an array containing the first 20 uncalibrated means. */
	public static float[] getUMeans() {
		return umeans;
	}

	/** Returns the ImageJ results table. */
	public static ResultsTable getResultsTable() {
		return systemRT;
	}

	/** Returns the number of digits displayed on the right of decimal point. */
	public static int getPrecision() {
		return precision;
	}

	/** Returns an updated Y coordinate based on
		the current "Invert Y Coordinates" flag. */
	public static int updateY(int y, int imageHeight) {
		if ((systemMeasurements&INVERT_Y)!=0)
			y = imageHeight-y-1;
		return y;
	}

	public void flush(){
		PrefSaveResult.save();
		textPanel = null;	imp = null;	header=null ;rt=null;


	}
	

}
	

